/*
 	Gasteroids v0.1 aplicacin didctica para dispositivos MIDP2.0
 	
    Copyright (C) 2006  Victor Carceler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package gpl.gasteroids;

import java.util.Vector;

import javax.microedition.lcdui.Graphics;

/**
 * Modela una nave y todos sus proyectiles.<p>
 * Las propiedades publicas de una nave son:
 * <ul>
 * 	<li>La posicin (x,y) en la que se encuentra
 * 	<li>El rea rectangular (x_min, y_min) y (x_max, y_max) en la que se puede mover
 * 	<li>La orientacin (arg) en radianes
 * 	<li>La velocidad (speed_x, speed_y)
 * 	<li>La fuerza a la que est sometida (force_x, force_y)
 * 	<li>La masa de la nave (m)
 * 	<li>La capacidad de los impulsores (thrust)
 * 	<li>El tiempo parcial durante el que ha mantenido los impulsores encendidos (engine_time)
 * 	<li>La longitud (L) de la nave
 * </ul>
 * 
 * <p>Una nave es responsable de los proyectiles que lanza. Estos proyectiles se guardan
 * en una instancia de {@link Vector}. Cuando la nave dispara, el proyectil sale con una
 * velocidad inicial resultante de la velocidad instantnea de la nave ms la proporcionada
 * por el can. 
 * 
 * @author Victor Carceler
 */
public class SpaceShip {
	public double x, y;							// Posicin X, Y
	public int x_min, x_max, y_min, y_max;		// Lmites a la posicin de la nave
	public double arg;							// Orientacin de la nave (radianes)
	public double speed_x=0.0d, speed_y=0.0d;	// Velocidad actual
	public double force_x=0.0d, force_y=0.0d; 	// Fuerza
	public double m = 0.75d;					// Masa de la nave
	public double thrust = 1.0d;				// Capacidad de los impulsores
	public double engine_time=0.0d;				// Tiempo con los impulsores encendidos
	public final int L = 15;					// Longitud de la nave
	private Vector bullets;						// Proyectiles disparados (y activos)
	private double bulletIS = 2.0d;				// Velocidad inicial de los disparos (adems de la que proporciona la nave)
	
	/**
	 * Constructor, inicializa la nave
	 */
	public SpaceShip() {
		bullets = new Vector();
	}
	
	/**
	 * Actualiza la velocidad de la nave. 
	 * <p>Para ello tiene en cuenta:
	 * <ul>
	 * 	<li>El tiempo que los motores han estado en marcha
	 * 	<li>El empuje de los motores
	 * 	<li>La orientacin de la nave
	 * 	<li>La velocidad a la que estaba viajando la nave
	 * </ul>
	 * <p>Cuando se actualiza la velocidad de la nave, se supone que la maniobra concluye.
	 * Por lo tanto el contador parcial de tiempo con los motores encendidos se pone a 0.
	 * @param t tiempo transcurrido en segundos
	 */
	public void updateSpeed(double t) {
		// Actualizamos la fuerza segn el tiempo que han estado activos los impulsores
		force_x = engine_time*thrust*Math.cos(arg);
		force_y = engine_time*thrust*Math.sin(arg);
		engine_time = 0.0d;
		
		// Actualizamos la velocidad
		speed_x += t*force_x/m;
		speed_y += t*force_y/m;
	}
	
	/**
	 * Actualiza la posicin de la nave suponiendo que han transcurrido t segundos.<p>
	 * Para ello:
	 * <ul>
	 * 	<li>Se calcula la nueva posicin que ha de ocupar la nave
	 * 	<li>Si la nave choca con un lmite no lo rebasa y la componente de velocidad que la impele a rebasarlo se pone a 0
	 * 	<li>Se calcula la nueva posicin que ha de ocupar cada uno de los proyectiles de la nave
	 * 	<li>Se decrementa en 1 el ttl de cada proyectil
	 * 	<li>Se descartan los proyectiles con ttl no superior a 0
	 * </ul>
	 * <p>Si se pretende que la nave rebote, o bien salga por un borde para entrar por otro (espacio continuo
	 * como el Asteroids original) este es el mtodo en el que picar cdigo.
	 * 
	 * @param t tiempo transcurrido en segundos
	 */
	public void updatePosition(double t) {
		// Actualizamos posicin de la nave
		x += speed_x * t;
		y += speed_y * t;
		if(x<=x_min) {x=x_min; speed_x = 0;}
		if(x>=x_max) {x=x_max; speed_x = 0;}
		if(y<=y_min) {y=y_min; speed_y = 0;} 
		if(y>=y_max) {y=y_max; speed_y = 0;}
		
		// Actualizamos posicin de los proyectiles
		for(int i=bullets.size()-1; i>=0; i--) {
			Bullet b = (Bullet)bullets.elementAt(i);
			b.x += b.speed_x * t;
			b.y += b.speed_y * t;
			b.ttl--;
			if(b.ttl <= 0) bullets.removeElementAt(i);
		}
	}
	
	/**
	 * Dibuja la nave y cada uno de los proyectiles.
	 * 
	 * @param g Objeto {@link Graphics} en el que dibujar
	 */
	public void paint(Graphics g)
	{
		// Dibujamos la nave
		g.setColor(0,0,0);
		g.drawArc((int)(x-L/2), (int)(y-L/2), L, L, 0, 360);
		g.drawLine((int)x, (int)y, (int)(x+L/2*Math.cos(arg)), (int)(y+L/2*Math.sin(arg)));
		
		// Dibujamos los proyectiles que ha lanzado
		for(int i=0; i<bullets.size(); i++) {
			Bullet b = (Bullet)bullets.elementAt(i);
			g.drawArc((int)(b.x)-2, (int)(b.y)-2, 4, 4, 0, 360);
			//System.out.println(b);
		}
		
		
		//System.out.println(this);
	}

	public String toString() {
		return "Nave -> Pos(" + x + "," + y + ") Speed(" + speed_x + "," + speed_y + ") Force(" + force_x +  "," + force_y + ")";
	}
	
	/**
	 * Pide a la nave que abra fuego.
	 * <p>Cada vez que se invoca a este mtodo, la nave lanza un nuevo proyectil.
	 * <p>El proyectil tiene una posicin determinada por la posicin en la que la nave dispar.
	 * <p>La velocidad del proyectil est determinada por la velocidad instantnea de la nave en el
	 * momento del disparo ms la velocidad proporcionada por su caon. Se tiene en cuenta la 
	 * orientacin de la nave en el momento del disparo.
	 *
	 */
	public void fire() {
		Bullet b = new Bullet();
		b.x = x; 										b.y = y;
		b.speed_x = bulletIS*Math.cos(arg)+speed_x;		b.speed_y = bulletIS*Math.sin(arg)+speed_y;
		b.ttl = 80;		// TTL por defecto para el proyectil
		bullets.addElement(b);
	}
	
	/**
	 * Retorna un {@link Vector} con los proyectiles controlados por la nave
	 * @return Vector con los proyectiles controlados por la nave
	 */
	public Vector getBullets() {
		return bullets;
	}
}
